@@ -53,39 +53,16 @@
 #include <fcntl.h>
 #include <errno.h>
 
-#ifdef __linux__
-#define DRM_VBLANK_RELATIVE 0x1;
-
-struct drm_wait_vblank_request {
-    int           type;
-    unsigned int  sequence;
-    unsigned long signal;
-};
-
-struct drm_wait_vblank_reply {
-    int          type;
-    unsigned int sequence;
-    long         tval_sec;
-    long         tval_usec;
-};
-
-typedef union drm_wait_vblank {
-    struct drm_wait_vblank_request request;
-    struct drm_wait_vblank_reply reply;
-} drm_wait_vblank_t;
-
-#define DRM_IOCTL_BASE                  'd'
-#define DRM_IOWR(nr,type)               _IOWR(DRM_IOCTL_BASE,nr,type)
-#define DRM_IOCTL_WAIT_VBLANK           DRM_IOWR(0x3a, drm_wait_vblank_t)
-
-#endif /* __linux__ */
+#ifdef HAVE_DRM
+#include <drm.h>
+#endif
 
 static void clutter_stage_window_iface_init (ClutterStageWindowIface *iface);
 
 static ClutterStageWindowIface *clutter_stage_glx_parent_iface = NULL;
 
 G_DEFINE_TYPE_WITH_CODE (ClutterStageGLX,
-                         clutter_stage_glx,
+                         _clutter_stage_glx,
                          CLUTTER_TYPE_STAGE_X11,
                          G_IMPLEMENT_INTERFACE (CLUTTER_TYPE_STAGE_WINDOW,
                                                 clutter_stage_window_iface_init));
@@ -132,6 +109,7 @@ clutter_stage_glx_realize (ClutterStageWindow *stage_window)
   ClutterBackendX11 *backend_x11;
   ClutterBackendGLX *backend_glx;
   ClutterBackend *backend;
+  int event_flags;
 
   CLUTTER_NOTE (ACTOR, "Realizing stage '%s' [%p]",
                 G_OBJECT_TYPE_NAME (stage_window),
@@ -218,43 +196,62 @@ clutter_stage_glx_realize (ClutterStageWindow *stage_window)
         }
     }
 
-  if (clutter_x11_has_event_retrieval ())
-    {
-      if (clutter_x11_has_xinput ())
-        {
-          XSelectInput (backend_x11->xdpy, stage_x11->xwin,
-                        StructureNotifyMask |
-                        FocusChangeMask |
-                        ExposureMask |
-                        KeyPressMask | KeyReleaseMask |
-                        EnterWindowMask | LeaveWindowMask |
-                        PropertyChangeMask);
+  /* the masks for the events we want to select on a stage window;
+   * KeyPressMask and KeyReleaseMask are necessary even with XI1
+   * because key events are broken with that extension, and will
+   * be fixed by XI2
+   */
+  event_flags = StructureNotifyMask
+              | FocusChangeMask
+              | ExposureMask
+              | PropertyChangeMask
+              | EnterWindowMask
+              | LeaveWindowMask
+              | KeyPressMask
+              | KeyReleaseMask;
+
+  /* if we don't use XI1 then we also want core pointer events */
+  if (!clutter_x11_has_xinput ())
+    event_flags |= (ButtonPressMask | ButtonReleaseMask | PointerMotionMask);
 #ifdef HAVE_XINPUT
-          _clutter_x11_select_events (stage_x11->xwin);
+  else
+    _clutter_x11_select_events (stage_x11->xwin);
 #endif
-        }
-      else
-        XSelectInput (backend_x11->xdpy, stage_x11->xwin,
-                      StructureNotifyMask |
-                      FocusChangeMask |
-                      ExposureMask |
-                      PointerMotionMask |
-                      KeyPressMask | KeyReleaseMask |
-                      ButtonPressMask | ButtonReleaseMask |
-                      EnterWindowMask | LeaveWindowMask |
-                      PropertyChangeMask);
+
+  /* we unconditionally select input events even with event retrieval
+   * disabled because we need to guarantee that the Clutter internal
+   * state is maintained when calling clutter_x11_handle_event() without
+   * requiring applications or embedding toolkits to select events
+   * themselves. if we did that, we'd have to document the events to be
+   * selected, and also update applications and embedding toolkits each
+   * time we added a new mask, or a new class of events.
+   *
+   * see: http://bugzilla.clutter-project.org/show_bug.cgi?id=998
+   * for the rationale of why we did conditional selection. it is now
+   * clear that a compositor should clear out the input region, since
+   * it cannot assume a perfectly clean slate coming from us.
+   *
+   * see: http://bugzilla.clutter-project.org/show_bug.cgi?id=2228
+   * for an example of things that break if we do conditional event
+   * selection.
+   */
+  XSelectInput (backend_x11->xdpy, stage_x11->xwin, event_flags);
 
 #ifdef GLX_INTEL_swap_event
-      if (clutter_feature_available (CLUTTER_FEATURE_SWAP_EVENTS))
-        {
-          GLXDrawable drawable =
-            stage_glx->glxwin ? stage_glx->glxwin : stage_x11->xwin;
-          glXSelectEvent (backend_x11->xdpy,
-                          drawable,
-                          GLX_BUFFER_SWAP_COMPLETE_INTEL_MASK);
-        }
-#endif /* GLX_INTEL_swap_event */
+  if (clutter_feature_available (CLUTTER_FEATURE_SWAP_EVENTS))
+    {
+      GLXDrawable drawable =
+        stage_glx->glxwin ? stage_glx->glxwin : stage_x11->xwin;
+
+      /* similarly to above, we unconditionally select this event
+       * because we rely on it to advance the master clock, and
+       * drive redraw/relayout, animations and event handling.
+       */
+      glXSelectEvent (backend_x11->xdpy,
+                      drawable,
+                      GLX_BUFFER_SWAP_COMPLETE_INTEL_MASK);
     }
+#endif /* GLX_INTEL_swap_event */
 
   /* no user resize.. */
   clutter_stage_x11_fix_window_size (stage_x11,
@@ -279,11 +276,11 @@ clutter_stage_glx_get_pending_swaps (ClutterStageWindow *stage_window)
 static void
 clutter_stage_glx_dispose (GObject *gobject)
 {
-  G_OBJECT_CLASS (clutter_stage_glx_parent_class)->dispose (gobject);
+  G_OBJECT_CLASS (_clutter_stage_glx_parent_class)->dispose (gobject);
 }
 
 static void
-clutter_stage_glx_class_init (ClutterStageGLXClass *klass)
+_clutter_stage_glx_class_init (ClutterStageGLXClass *klass)
 {
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 
@@ -291,7 +288,7 @@ clutter_stage_glx_class_init (ClutterStageGLXClass *klass)
 }
 
 static void
-clutter_stage_glx_init (ClutterStageGLX *stage)
+_clutter_stage_glx_init (ClutterStageGLX *stage)
 {
 }
 
@@ -383,12 +380,11 @@ clutter_stage_glx_add_redraw_clip (ClutterStageWindow *stage_window,
   if (stage_clip == NULL)
     {
       stage_glx->bounding_redraw_clip.width = 0;
+      stage_glx->initialized_redraw_clip = TRUE;
       return;
     }
 
-  /* Do nothing on an empty clip to avoid confusing with out magic-flag
-   * degenerate clip
-   */
+  /* Ignore requests to add degenerate/empty clip rectangles */
   if (stage_clip->width == 0 || stage_clip->height == 0)
     return;
 
@@ -405,13 +401,6 @@ clutter_stage_glx_add_redraw_clip (ClutterStageWindow *stage_window,
 			      &stage_glx->bounding_redraw_clip);
     }
 
-  /* FIXME: This threshold was plucked out of thin air! */
-  if (stage_glx->bounding_redraw_clip.height > 300)
-    {
-      /* Set a degenerate clip to force a full redraw */
-      stage_glx->bounding_redraw_clip.width = 0;
-    }
-
 #if 0
   redraw_area = (stage_glx->bounding_redraw_clip.width *
                  stage_glx->bounding_redraw_clip.height);
@@ -449,7 +438,7 @@ clutter_stage_window_iface_init (ClutterStageWindowIface *iface)
   /* the rest is inherited from ClutterStageX11 */
 }
 
-#ifdef __linux__
+#ifdef HAVE_DRM
 static int
 drm_wait_vblank(int fd, drm_wait_vblank_t *vbl)
 {
@@ -458,78 +447,49 @@ drm_wait_vblank(int fd, drm_wait_vblank_t *vbl)
     do
       {
         ret = ioctl(fd, DRM_IOCTL_WAIT_VBLANK, vbl);
-        vbl->request.type &= ~DRM_VBLANK_RELATIVE;
+        vbl->request.type &= ~_DRM_VBLANK_RELATIVE;
         rc = errno;
       }
     while (ret && rc == EINTR);
 
     return rc;
 }
-#endif /* __linux__ */
+#endif /* HAVE_DRM */
 
 static void
-glx_wait_for_vblank (ClutterBackendGLX *backend_glx)
+wait_for_vblank (ClutterBackendGLX *backend_glx)
 {
-  /* If we are going to wait for VBLANK manually, we not only need
-   * to flush out pending drawing to the GPU before we sleep, we
-   * need to wait for it to finish. Otherwise, we may end up with
-   * the situation:
-   *
-   *        - We finish drawing      - GPU drawing continues
-   *        - We go to sleep         - GPU drawing continues
-   * VBLANK - We call glXSwapBuffers - GPU drawing continues
-   *                                 - GPU drawing continues
-   *                                 - Swap buffers happens
-   *
-   * Producing a tear. Calling glFinish() first will cause us to properly
-   * wait for the next VBLANK before we swap. This obviously does not
-   * happen when we use GLX_SWAP and let the driver do the right thing
-   */
+  if (backend_glx->vblank_type == CLUTTER_VBLANK_NONE)
+    return;
 
-  switch (backend_glx->vblank_type)
+  if (backend_glx->wait_video_sync)
     {
-    case CLUTTER_VBLANK_GLX_SWAP:
-      CLUTTER_NOTE (BACKEND, "Waiting for vblank (swap)");
-      break;
+      unsigned int retraceCount;
 
-    case CLUTTER_VBLANK_GLX:
-      {
-        unsigned int retraceCount;
+      CLUTTER_NOTE (BACKEND, "Waiting for vblank (wait_video_sync)");
+      backend_glx->get_video_sync (&retraceCount);
+      backend_glx->wait_video_sync (2,
+                                    (retraceCount + 1) % 2,
+                                    &retraceCount);
+    }
+  else
+    {
+#ifdef HAVE_DRM
+      drm_wait_vblank_t blank;
 
-        glFinish ();
-
-        CLUTTER_NOTE (BACKEND, "Waiting for vblank (wait_video_sync)");
-        backend_glx->get_video_sync (&retraceCount);
-        backend_glx->wait_video_sync (2,
-                                      (retraceCount + 1) % 2,
-                                      &retraceCount);
-      }
-      break;
-
-    case CLUTTER_VBLANK_DRI:
-#ifdef __linux__
-      {
-        drm_wait_vblank_t blank;
-
-        glFinish ();
-
-        CLUTTER_NOTE (BACKEND, "Waiting for vblank (drm)");
-        blank.request.type     = DRM_VBLANK_RELATIVE;
-        blank.request.sequence = 1;
-        blank.request.signal   = 0;
-        drm_wait_vblank (backend_glx->dri_fd, &blank);
-      }
-#endif
-      break;
-
-    case CLUTTER_VBLANK_NONE:
-    default:
-      break;
+      CLUTTER_NOTE (BACKEND, "Waiting for vblank (drm)");
+      blank.request.type     = _DRM_VBLANK_RELATIVE;
+      blank.request.sequence = 1;
+      blank.request.signal   = 0;
+      drm_wait_vblank (backend_glx->dri_fd, &blank);
+#else
+      CLUTTER_NOTE (BACKEND, "No vblank mechanism found");
+#endif /* HAVE_DRM */
     }
 }
 
 void
-clutter_stage_glx_redraw (ClutterStageGLX *stage_glx,
+_clutter_stage_glx_redraw (ClutterStageGLX *stage_glx,
                           ClutterStage *stage)
 {
   ClutterBackend    *backend;
@@ -537,6 +497,7 @@ clutter_stage_glx_redraw (ClutterStageGLX *stage_glx,
   ClutterBackendGLX *backend_glx;
   ClutterStageX11   *stage_x11;
   GLXDrawable        drawable;
+  unsigned int       video_sync_count;
   CLUTTER_STATIC_TIMER (painting_timer,
                         "Redrawing", /* parent */
                         "Painting actors",
@@ -547,10 +508,10 @@ clutter_stage_glx_redraw (ClutterStageGLX *stage_glx,
                         "glXSwapBuffers",
                         "The time spent blocked by glXSwapBuffers",
                         0 /* no application private data */);
-  CLUTTER_STATIC_TIMER (copy_sub_buffer_timer,
+  CLUTTER_STATIC_TIMER (blit_sub_buffer_timer,
                         "Redrawing", /* parent */
-                        "glXCopySubBufferMESA",
-                        "The time spent blocked by glXCopySubBufferMESA",
+                        "glx_blit_sub_buffer",
+                        "The time spent in _glx_blit_sub_buffer",
                         0 /* no application private data */);
 
   backend     = clutter_get_default_backend ();
@@ -561,7 +522,7 @@ clutter_stage_glx_redraw (ClutterStageGLX *stage_glx,
 
   CLUTTER_TIMER_START (_clutter_uprof_context, painting_timer);
 
-  if (backend_glx->copy_sub_buffer &&
+  if (backend_glx->can_blit_sub_buffer &&
       /* NB: a degenerate redraw clip width == full stage redraw */
       (stage_glx->bounding_redraw_clip.width != 0) &&
       G_LIKELY (!(clutter_paint_debug_flags &
@@ -585,11 +546,19 @@ clutter_stage_glx_redraw (ClutterStageGLX *stage_glx,
 
   drawable = stage_glx->glxwin ? stage_glx->glxwin : stage_x11->xwin;
 
-  /* wait for the next vblank */
-  glx_wait_for_vblank (CLUTTER_BACKEND_GLX (backend));
+  /* If we might ever use _clutter_backend_glx_blit_sub_buffer then we
+   * always need to keep track of the video_sync_count so that we can
+   * throttle blits.
+   *
+   * Note: we get the count *before* we issue any glXCopySubBuffer or
+   * blit_sub_buffer request in case the count would go up before
+   * returning control to us.
+   */
+  if (backend_glx->can_blit_sub_buffer && backend_glx->get_video_sync)
+    backend_glx->get_video_sync (&video_sync_count);
 
   /* push on the screen */
-  if (backend_glx->copy_sub_buffer &&
+  if (backend_glx->can_blit_sub_buffer &&
       /* NB: a degenerate redraw clip width == full stage redraw */
       (stage_glx->bounding_redraw_clip.width != 0) &&
       G_LIKELY (!(clutter_paint_debug_flags &
@@ -599,11 +568,9 @@ clutter_stage_glx_redraw (ClutterStageGLX *stage_glx,
       ClutterGeometry copy_area;
 
       CLUTTER_NOTE (BACKEND,
-                    "glXCopySubBufferMESA (display: %p, "
-                                          "window: 0x%lx, "
+                    "_glx_blit_sub_buffer (window: 0x%lx, "
                                           "x: %d, y: %d, "
                                           "width: %d, height: %d)",
-                    backend_x11->xdpy,
                     (unsigned long) drawable,
                     stage_glx->bounding_redraw_clip.x,
                     stage_glx->bounding_redraw_clip.y,
@@ -660,14 +627,51 @@ clutter_stage_glx_redraw (ClutterStageGLX *stage_glx,
       copy_area.width = clip->width;
       copy_area.height = clip->height;
 
-      CLUTTER_TIMER_START (_clutter_uprof_context, copy_sub_buffer_timer);
-      backend_glx->copy_sub_buffer (backend_x11->xdpy,
-                                    drawable,
-                                    copy_area.x,
-                                    copy_area.y,
-                                    copy_area.width,
-                                    copy_area.height);
-      CLUTTER_TIMER_STOP (_clutter_uprof_context, copy_sub_buffer_timer);
+      /* glXCopySubBufferMESA and glBlitFramebuffer are not integrated
+       * with the glXSwapIntervalSGI mechanism which we usually use to
+       * throttle the Clutter framerate to the vertical refresh and so
+       * we have to manually wait for the vblank period...
+       */
+
+      /* Here 'is_synchronized' only means that the blit won't cause a
+       * tear, ie it won't prevent multiple blits per retrace if they
+       * can all be performed in the blanking period. If that's the
+       * case then we still want to use the vblank sync menchanism but
+       * we only need it to throttle redraws.
+       */
+      if (!backend_glx->blit_sub_buffer_is_synchronized)
+        {
+          /* XXX: note that glXCopySubBuffer, at least for Intel, is
+           * synchronized with the vblank but glBlitFramebuffer may
+           * not be so we use the same scheme we do when calling
+           * glXSwapBuffers without the swap_control extension and
+           * call glFinish () before waiting for the vblank period.
+           *
+           * See where we call glXSwapBuffers for more details.
+           */
+          glFinish ();
+          wait_for_vblank (CLUTTER_BACKEND_GLX (backend));
+        }
+      else if (backend_glx->get_video_sync)
+        {
+          /* If we have the GLX_SGI_video_sync extension then we can
+           * be a bit smarter about how we throttle blits by avoiding
+           * any waits if we can see that the video sync count has
+           * already progressed. */
+          if (backend_glx->last_video_sync_count == video_sync_count)
+            wait_for_vblank (CLUTTER_BACKEND_GLX (backend));
+        }
+      else
+        wait_for_vblank (CLUTTER_BACKEND_GLX (backend));
+
+      CLUTTER_TIMER_START (_clutter_uprof_context, blit_sub_buffer_timer);
+      _clutter_backend_glx_blit_sub_buffer (backend_glx,
+                                            drawable,
+                                            copy_area.x,
+                                            copy_area.y,
+                                            copy_area.width,
+                                            copy_area.height);
+      CLUTTER_TIMER_STOP (_clutter_uprof_context, blit_sub_buffer_timer);
     }
   else
     {
@@ -681,11 +685,37 @@ clutter_stage_glx_redraw (ClutterStageGLX *stage_glx,
       if (clutter_feature_available (CLUTTER_FEATURE_SWAP_EVENTS))
         stage_glx->pending_swaps++;
 
+      if (backend_glx->vblank_type != CLUTTER_VBLANK_GLX_SWAP &&
+          backend_glx->vblank_type != CLUTTER_VBLANK_NONE)
+        {
+          /* If we are going to wait for VBLANK manually, we not only
+           * need to flush out pending drawing to the GPU before we
+           * sleep, we need to wait for it to finish. Otherwise, we
+           * may end up with the situation:
+           *
+           *        - We finish drawing      - GPU drawing continues
+           *        - We go to sleep         - GPU drawing continues
+           * VBLANK - We call glXSwapBuffers - GPU drawing continues
+           *                                 - GPU drawing continues
+           *                                 - Swap buffers happens
+           *
+           * Producing a tear. Calling glFinish() first will cause us
+           * to properly wait for the next VBLANK before we swap. This
+           * obviously does not happen when we use _GLX_SWAP and let
+           * the driver do the right thing
+           */
+          glFinish ();
+
+          wait_for_vblank (CLUTTER_BACKEND_GLX (backend));
+        }
+
       CLUTTER_TIMER_START (_clutter_uprof_context, swapbuffers_timer);
       glXSwapBuffers (backend_x11->xdpy, drawable);
       CLUTTER_TIMER_STOP (_clutter_uprof_context, swapbuffers_timer);
     }
 
+  backend_glx->last_video_sync_count = video_sync_count;
+
   /* reset the redraw clipping for the next paint... */
   stage_glx->initialized_redraw_clip = FALSE;
 }
